theory Simplex_Ocaml_Interface
  imports Ord.IA_Checker
    Auxx.Shows_Literal_List
    RA_Checker
begin


datatype 'a Formula =
  FAtom "'a"
| FNegate "'a Formula"
| FConjunction "'a Formula list"
| FDisjunction "'a Formula list"

fun of_Formula :: "'a Formula \<Rightarrow> 'a formula" where
  "of_Formula (FAtom a) = Atom a" 
| "of_Formula (FNegate \<phi>) = (\<not>\<^sub>f (of_Formula \<phi>))" 
| "of_Formula (FConjunction \<phi>s) = form_all (map of_Formula \<phi>s)" 
| "of_Formula (FDisjunction \<phi>s) = form_ex (map of_Formula \<phi>s)" 

context formula
begin
fun Formula :: "'atom Formula \<Rightarrow> bool" where
  "Formula (FAtom a) = atom a"
| "Formula (FNegate \<phi>) = Formula \<phi>"
| "Formula (FConjunction \<phi>s) = (\<forall>\<phi> \<in> set \<phi>s. Formula \<phi>)"
| "Formula (FDisjunction \<phi>s) = (\<forall>\<phi> \<in> set \<phi>s. Formula \<phi>)"

lemma formula_of_Formula: "formula (of_Formula \<phi>) = Formula \<phi>" 
  by (induct \<phi>, auto)

context fixes \<alpha> :: "'assignment" begin

fun Satisfies :: "'atom Formula \<Rightarrow> bool" where
  "Satisfies (FAtom a) \<longleftrightarrow> eval_true \<alpha> a"
| "Satisfies (FNegate \<phi>) \<longleftrightarrow> \<not> Satisfies \<phi>"
| "Satisfies (FConjunction \<phi>s) \<longleftrightarrow> (\<forall>\<phi> \<in> set \<phi>s. Satisfies \<phi>)"
| "Satisfies (FDisjunction \<phi>s) \<longleftrightarrow> (\<exists>\<phi> \<in> set \<phi>s. Satisfies \<phi>)"

lemma satisfies_of_Formula: "satisfies \<alpha> (of_Formula \<phi>) \<longleftrightarrow> Satisfies \<phi>" 
  by (induct \<phi>, auto)
end

declare formula.Formula.simps[code]

definition Valid where
  "(Valid \<phi>) = (\<forall> \<alpha>. assignment \<alpha> \<longrightarrow> Satisfies \<alpha> \<phi>)"

lemma valid_of_Formula: "valid (of_Formula \<phi>) \<longleftrightarrow> Valid \<phi>"  
  unfolding valid_def Valid_def satisfies_of_Formula ..
end

instantiation integer :: "showl" 
begin
definition "showsl_integer == showsl o int_of_integer" 
definition "showsl_list_integer == showsl_list o map int_of_integer" 

instance ..
end
  

definition check_valid_IA_formula :: "(IA.sig, _ \<times> IA.ty) Term.term Formula \<Rightarrow> _ + unit" where
  "check_valid_IA_formula \<phi> = map_sum (\<lambda> s. s (STR '''')) id (IA.check_formula hint.Default (of_Formula \<phi>))" 

definition check_valid_RA_formula :: "(RA.sig, _ \<times> RA.ty) Term.term Formula \<Rightarrow> _ + unit" where
  "check_valid_RA_formula \<phi> = map_sum (\<lambda> s. s (STR '''')) id (RA.check_formula hint.Default (of_Formula \<phi>))" 

lemma isOK_map_sum: "isOK (map_sum f g x) = isOK x" by (cases x, auto)

lemma check_valid_IA_formula: assumes ok: "isOK(check_valid_IA_formula \<phi>)" 
  and "IA.Formula \<phi>" 
  shows "IA.Valid \<phi>" 
proof -
  from IA.check_formula[OF ok[unfolded check_valid_IA_formula_def isOK_map_sum], unfolded IA.valid_of_Formula
      IA.formula_of_Formula]
  have "IA.Formula \<phi> \<Longrightarrow> IA.Valid \<phi>" .
  with assms show "IA.Valid \<phi>" by blast
qed 

lemma check_valid_RA_formula: assumes ok: "isOK(check_valid_RA_formula \<phi>)" 
  and "RA.Formula \<phi>" 
  shows "RA.Valid \<phi>" 
proof -
  from RA.check_formula[OF ok[unfolded check_valid_RA_formula_def isOK_map_sum], unfolded RA.valid_of_Formula
      RA.formula_of_Formula]
  have "RA.Formula \<phi> \<Longrightarrow> RA.Valid \<phi>" .
  with assms show "RA.Valid \<phi>" by blast
qed 

definition check_well_formed_IA_formula_integer_vars :: "(IA.sig, integer \<times> IA.ty) Term.term Formula \<Rightarrow> bool" where
  "check_well_formed_IA_formula_integer_vars = IA.Formula" 

definition check_well_formed_IA_formula_string_vars :: "(IA.sig, string \<times> IA.ty) Term.term Formula \<Rightarrow> bool" where
  "check_well_formed_IA_formula_string_vars = IA.Formula" 

definition check_valid_IA_formula_integer_vars :: "(IA.sig, integer \<times> IA.ty) Term.term Formula \<Rightarrow> _ + unit" where
  "check_valid_IA_formula_integer_vars = check_valid_IA_formula"

definition check_valid_IA_formula_string_vars :: "(IA.sig, string \<times> IA.ty) Term.term Formula \<Rightarrow> _ + unit" where
  "check_valid_IA_formula_string_vars = check_valid_IA_formula"

definition check_well_formed_RA_formula_integer_vars :: "(RA.sig, integer \<times> RA.ty) Term.term Formula \<Rightarrow> bool" where
  "check_well_formed_RA_formula_integer_vars = RA.Formula" 

definition check_well_formed_RA_formula_string_vars :: "(RA.sig, string \<times> RA.ty) Term.term Formula \<Rightarrow> bool" where
  "check_well_formed_RA_formula_string_vars = RA.Formula" 

definition check_valid_RA_formula_integer_vars :: "(RA.sig, integer \<times> RA.ty) Term.term Formula \<Rightarrow> _ + unit" where
  "check_valid_RA_formula_integer_vars = check_valid_RA_formula"

definition check_valid_RA_formula_string_vars :: "(RA.sig, string \<times> RA.ty) Term.term Formula \<Rightarrow> _ + unit" where
  "check_valid_RA_formula_string_vars = check_valid_RA_formula"
 
definition create_rat :: "integer \<Rightarrow> integer \<Rightarrow> rat" where
  "create_rat num denom = ((of_int o int_of_integer) num / (of_int o int_of_integer) denom)" 

export_code 
  check_valid_IA_formula_integer_vars check_valid_IA_formula_string_vars
  check_well_formed_IA_formula_integer_vars check_well_formed_IA_formula_string_vars
  IA_locale.LessF IA_locale.BoolT 
  check_valid_RA_formula_integer_vars check_valid_RA_formula_string_vars
  check_well_formed_RA_formula_integer_vars check_well_formed_RA_formula_string_vars
  RA_locale.LessF RA_locale.BoolT 
  FAtom Var 
  int_of_integer nat_of_integer String.explode
  create_rat
  Inl
  in OCaml module_name Simplex_Validity_Checker file "~/Code/tmp/simplex.ml" 


end
